#include "viewer.h"

using namespace std;
using namespace qglviewer;

//************************************************************
// Implementation of Viewer
//************************************************************
Viewer::Viewer(QWidget* parent)
  : QGLViewer(parent),
    showSilhouettes(false),
    showShadowQuads(false),
    showShadows(false)
{
}
void
Viewer::loadModel(const char* fileName)
{
  OpenMesh::IO::Options ropt;
  if (!OpenMesh::IO::read_mesh(mesh,fileName,ropt))
  {
    cerr<<"Error loading mesh from file "<<fileName<<endl;
    return;
  }
  cout<<"OpenMesh imported ["<<fileName<<"]"<<endl;
  mesh.request_face_normals();
  mesh.request_vertex_normals();
  mesh.update_normals();
  // Compute bounding box
  bbox.clear();
  for (Mesh::VertexIter v=mesh.vertices_begin();v!=mesh.vertices_end();++v)
  {
    bbox.addPoint(mesh.point(v));
  }  
  setSceneCenter(bbox.center());
  setSceneRadius(bbox.sizeMax()/2.0f);
  showEntireScene();
}
void
Viewer::init()
{
  light = new qglviewer::ManipulatedFrame();
  light->setPosition(bbox.center());
  connect(light,SIGNAL(modified()),this,SLOT(updateGL()));
  setMouseTracking(true);
  glClearColor(0.498039f,0.498039f,0.7490f,1.0f);
}
void
Viewer::preDraw()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
  camera()->loadProjectionMatrix();
  camera()->loadModelViewMatrix();
}
void
Viewer::draw()
{
  glPushAttrib(GL_ALL_ATTRIB_BITS);
  //////////////////////////////////////////////////////////////
  // Draw the light
  //////////////////////////////////////////////////////////////
  Vec L = light->position();
  float Lw[4] = { L[0],L[1],L[2],1.0f };
  glLightfv(GL_LIGHT0,GL_POSITION,Lw);

  glPointSize(light->grabsMouse()?12.0f:8.0f);
  glDisable(GL_LIGHTING);
  glBegin(GL_POINTS);
  glColor3f(1.0f,1.0f,0.0f);
  glVertex3fv(L);
  glEnd();
  //////////////////////////////////////////////////////////////
  // Draw the model
  //////////////////////////////////////////////////////////////
  if (showShadows)
  {
    glEnable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    glColor3f(1.0f,1.0f,1.0f);
    renderModel();
    //////////////////////////////////////////////////////////////
    // Shadow volumes
    //////////////////////////////////////////////////////////////  
    glColorMask(false,false,false,false);
    glDepthMask(false);
    glDepthFunc(GL_LESS);
    glEnable(GL_STENCIL_TEST);
    glEnable(GL_CULL_FACE);
    //------------------------------------------------------------
    // Z-Pass
    //------------------------------------------------------------
    // Front facing shadow quads will increment stencil
    glStencilFunc(GL_ALWAYS,0,~0);
    glStencilOp(GL_KEEP,GL_KEEP,GL_INCR);
    glStencilMask(~0);
    glCullFace(GL_BACK);
    renderShadowQuads();
    // Back facing shadow quads will decrement stencil
    glStencilFunc(GL_ALWAYS,0,~0);
    glStencilOp(GL_KEEP,GL_KEEP,GL_DECR);
    glStencilMask(~0);
    glCullFace(GL_FRONT);
    renderShadowQuads();
    //////////////////////////////////////////////////////////////
    // Render scene to get diffuse lighting where unshadowed
    //////////////////////////////////////////////////////////////
    glColorMask(true,true,true,true);
    glDepthFunc(GL_EQUAL);
    glEnable(GL_LIGHT0);
    glEnable(GL_DEPTH_TEST);
    glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
    glStencilFunc(GL_EQUAL,0,~0);
    glDisable(GL_CULL_FACE);
    glEnable(GL_LIGHT0);
    renderModel();    
  }
  else
  {
    glEnable(GL_LIGHTING);
    glColor3f(1.0f,1.0f,1.0f);
    renderModel();
  }
  //////////////////////////////////////////////////////////////
  // Draw the silhouette edges
  //////////////////////////////////////////////////////////////
  if (showSilhouettes)
  {
    glDisable(GL_LIGHTING);
    glDepthRange(0.0f,0.999f);
    glColor3f(1.0f,0.0f,0.0f);
    glLineWidth(3.0f);
    renderSilhouettes();
    glDepthRange(0.0f,1.0f);
  }
  //////////////////////////////////////////////////////////////
  // Draw the shadow quads
  //////////////////////////////////////////////////////////////
  if (showShadowQuads)
  {
    glDisable(GL_LIGHTING);
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
    static const float color[3] = { 1.0f,1.0f,0.0f };
    static const float a = 0.5f;
    glColor4f(color[0]*a,color[1]*a,color[2]*a,a);
    renderShadowQuads();
    glDisable(GL_BLEND);
  }
  //
  glPopAttrib();
}
void
Viewer::renderModel() const
{
  glBegin(GL_TRIANGLES);
  for (Mesh::ConstFaceIter f=mesh.faces_begin();f!=mesh.faces_end();++f)
  {
    glNormal3fv(mesh.normal(f));
    for (Mesh::ConstFaceVertexIter v=mesh.cfv_iter(f);v;++v)
    {
      glVertex3fv(mesh.point(v));
    }
  }
  glEnd();
}
void
Viewer::renderSilhouettes() const
{
  Vec L = light->position();
  glBegin(GL_LINES);
  for (Mesh::ConstEdgeIter e=mesh.edges_begin();e!=mesh.edges_end();++e)
  {
    Mesh::HalfedgeHandle h0 = mesh.halfedge_handle(e,0);
    Mesh::HalfedgeHandle h1 = mesh.halfedge_handle(e,1);
    Vec F(mesh.point(mesh.from_vertex_handle(h0)));
    Vec T(mesh.point(mesh.from_vertex_handle(h1)));
    Vec n0(mesh.normal(mesh.face_handle(h0)));
    Vec n1(mesh.normal(mesh.face_handle(h1)));
    float ps0 = ((F-L)*n0);
    float ps1 = ((F-L)*n1);
    if (ps0*ps1 < 0)
    {
      glVertex3fv(F);
      glVertex3fv(T);
    }
  }  
  glEnd();
}
void
Viewer::renderShadowQuads() const
{
  Vec L = light->position();
  glBegin(GL_QUADS);
  for (Mesh::ConstEdgeIter e=mesh.edges_begin();e!=mesh.edges_end();++e)
  {
    Mesh::HalfedgeHandle h0 = mesh.halfedge_handle(e,0);
    Mesh::HalfedgeHandle h1 = mesh.halfedge_handle(e,1);
    Vec F(mesh.point(mesh.from_vertex_handle(h0)));
    Vec T(mesh.point(mesh.from_vertex_handle(h1)));
    Vec n0(mesh.normal(mesh.face_handle(h0)));
    Vec n1(mesh.normal(mesh.face_handle(h1)));
    float ps0 = ((F-L)*n0);
    float ps1 = ((F-L)*n1);
    if (ps0*ps1 < 0)
    {
      if (ps0 < 0)
      {
	glVertex3fv(T);
	glVertex3fv(F);
	glVertex4f(F[0]-L[0],F[1]-L[1],F[2]-L[2],0.0f);
	glVertex4f(T[0]-L[0],T[1]-L[1],T[2]-L[2],0.0f);
      }
      else
      {
	glVertex4f(T[0]-L[0],T[1]-L[1],T[2]-L[2],0.0f);
	glVertex4f(F[0]-L[0],F[1]-L[1],F[2]-L[2],0.0f);
	glVertex3fv(F);
	glVertex3fv(T);
      }
    }
  }
  glEnd();
}
void
Viewer::keyPressEvent(QKeyEvent* e)
{
  switch (e->key())
  {
  case Key_R:
    restoreFromFile();
    updateGL();
    break;
  case Key_E: showSilhouettes = !showSilhouettes;updateGL();break;
  case Key_Q: showShadowQuads = !showShadowQuads;updateGL();break;
  case Key_A: showShadows = !showShadows;updateGL();break;
  default:
    QGLViewer::keyPressEvent(e);
  }
}
